import java.io.*;

public class Codegen {

	private StringBuffer staticBuffer       = new StringBuffer(); // fuer statische Variablen
	private StringBuffer clinitBuffer       = new StringBuffer(); // fuer den Klassenkonstruktor
	private StringBuffer methodeBuffer      = new StringBuffer(); // fuer die jeweils aktuelle Methode
	private StringBuffer multiMethodeBuffer = new StringBuffer(); // fuer alle bisher abgeschlossenen Methoden
	private StringBuffer mainBuffer         = new StringBuffer(); // zum Zusammenbasteln aller Module

	private int methodeBufferAddress;                             // notwendig um .limit und .local nachzutragen
	private int methodeCurrentStack         = 0;                  // aktuelle Stackgroesse pro Methode
	private int methodeMaximumStack         = 0;                  // maximale Stackgroesse pro Methode
	private int methodeNumberOfLocals       = 0;                  // maximale Anzahl von lokalen Variablen pro Methode

	private int clinitBufferAddress;                              // notwendig um .limit und .local nachzutragen
	private int clinitCurrentStack          = 0;                  // aktuelle Stackgroesse im Klassenkonstruktor
	private int clinitMaximumStack          = 0;                  // maximale Stackgroesse im Klassenkonstruktor

	private int labelcounter                = 1;                  // Nummer des naechsten Labels
	private boolean global                  = true;               // sind wir gerade global (true) oder in einer Methode (false)?

	private File inputFile;                                       // Eingabefile
	private File outputFile;                                      // Ausgabefile
	private String className;                                     // Name der zu generierenden Klasse
	
	public String init(File file) {
		
		// Initialisieren von File- und Klassennamen
		if (file == null) {
			inputFile  = null;
			outputFile = new File("MinaOut.jas");
			className = "MinaOut";
		}
		else {
			inputFile = file;
			try {
				if (!file.getName().toLowerCase().endsWith(".c0")) Minako.error("Wrong InputFileName -- \"*.c0\" expected");
				outputFile = new File(file.getCanonicalPath().substring(0, file.getCanonicalPath().length()-3) + ".jas");
				className = inputFile.getName().substring(0, inputFile.getName().length() - 3);
				System.err.println("Writing to " + outputFile);
			} catch (Exception e) {
				System.err.println("Exception while trying to set outputFile\n" + e);
				System.exit(-1);
			}
		}

		// Erstellen der Kopfzeilen
		mainBuffer.append("; ------- Header --------------------------------------------\n");
		if (inputFile != null) mainBuffer.append(".source " + inputFile.getName() + '\n');
		mainBuffer.append(".class  public synchronized " + className + '\n');
		mainBuffer.append(".super  java/lang/Object\n");
		mainBuffer.append("; -----------------------------------------------------------\n\n");

		// Erstellen des Objektkonstruktors (konstant)
		multiMethodeBuffer.append("; ------- Constructor ---------------------------------------\n");
		multiMethodeBuffer.append(".method public <init>()V\n");
		multiMethodeBuffer.append("\t.limit stack 1\n");
		multiMethodeBuffer.append("\t.limit locals 1\n");
		multiMethodeBuffer.append("\taload_0\n");
		multiMethodeBuffer.append("\tinvokenonvirtual java/lang/Object/<init>()V\n");
		multiMethodeBuffer.append("\treturn\n");
		multiMethodeBuffer.append(".end method\n");
		multiMethodeBuffer.append("; -----------------------------------------------------------\n\n");

		// Vorbereiten des Bereiches fuer globale Variablen und des Klassenkonstruktors
		staticBuffer.append("; ------- Class Variables -----------------------------------\n");

		clinitBuffer.append("; ------- Class Constructor ---------------------------------\n");
		clinitBuffer.append(".method static <clinit>()V\n");
		clinitBufferAddress = clinitBuffer.length();
		clinitBuffer.append("\t.limit locals 0\n");

		return className;
	}

	public void finishCode() {

		// Schliessen der Bloecke fuer globale Variablen und Klassenkonstruktor
		staticBuffer.append("; -----------------------------------------------------------\n\n");

		clinitBuffer.insert(clinitBufferAddress, "\t.limit stack " + clinitMaximumStack + '\n');
		clinitBuffer.append("\treturn\n");
		clinitBuffer.append(".end method\n");
		clinitBuffer.append("; -----------------------------------------------------------\n\n");

		// Zusammenfuegen des Programmes
		mainBuffer.append(staticBuffer);
		mainBuffer.append(multiMethodeBuffer);
		mainBuffer.append(clinitBuffer);
		
		try {
			BufferedWriter bufOut = new BufferedWriter(new FileWriter(outputFile));
			bufOut.write(mainBuffer.toString());
			bufOut.close();
		}
		catch (Exception e) {
			System.err.println("Exception while trying to write to file \"" + outputFile + "\":");
			System.err.println(e);
			System.exit(-1);
		}

		System.out.println(mainBuffer);
	}

	public void enterFunction(String name, int typ, String paramString) {
		methodeBuffer = new StringBuffer();
		methodeCurrentStack = 0;
		methodeMaximumStack = 0;
		labelcounter = 1;
		global = false;

		if (name.equals("main")) {
			if (typ != Minako.TYP_VOID) Minako.error("main has to be type void");
			if (paramString.length() != 0) Minako.error("main is'nt allowed to have parameters");
			methodeNumberOfLocals = 1;
			methodeBuffer.append("; ------- Methode ---- void main() --------------------------\n");
			methodeBuffer.append(".method public static main([Ljava/lang/String;)V\n");
		}
		else {
			methodeNumberOfLocals = 0;
			for (int i=0; i<paramString.length(); i++) methodeNumberOfLocals++;
			methodeBuffer.append("; ------- Methode -------------------------------------------\n");
			if (typ == Minako.TYP_INT) methodeBuffer.append(".method public static " + name + "(" + paramString + ")I\n");
			else if (typ == Minako.TYP_FLOAT) methodeBuffer.append(".method public static " + name + "(" + paramString + ")F\n");
			else if (typ == Minako.TYP_BOOLEAN) methodeBuffer.append(".method public static " + name + "(" + paramString + ")Z\n");
			else if (typ == Minako.TYP_VOID) methodeBuffer.append(".method public static " + name + "(" + paramString + ")V\n");
			else assert(false);
		}

		methodeBufferAddress = methodeBuffer.length();
	}
	
	public void leaveFunction() {
		global = true;

		methodeBuffer.insert(methodeBufferAddress, "\t.limit locals " + methodeNumberOfLocals + '\n');
		methodeBuffer.insert(methodeBufferAddress, "\t.limit stack " + methodeMaximumStack + '\n');
		methodeBuffer.append("\treturn\n");
		methodeBuffer.append(".end method\n");

		multiMethodeBuffer.append(methodeBuffer);
		multiMethodeBuffer.append("; -----------------------------------------------------------\n\n");
	}

	public int getNextLabel() {
		return labelcounter++;
	}

	public int getAddress() {
		if (global) return clinitBuffer.length();
		else return methodeBuffer.length();
	}

	public void appendCommand(String cmd) {
		appendCommand(cmd, 0);
	}

	public void appendCommand(String cmd, int stackdiff) {
		if (global) clinitBuffer.append('\t' + cmd + '\n');
		else methodeBuffer.append('\t' + cmd + '\n');
		increaseStackBy(stackdiff);
	}

	public void appendJump(String cmd, int label) {
		if (global) clinitBuffer.append('\t' + cmd + " Label" + label + '\n');
		else methodeBuffer.append('\t' + cmd + " Label" + label + '\n');
	}

	public void insertCommand(int address, String cmd) {
		if (global) clinitBuffer.insert(address, '\t' + cmd + '\n');
		else methodeBuffer.insert(address, '\t' + cmd + '\n');
	}

	public void appendLabel(int label) {
		if (global) clinitBuffer.append("Label" + label + ":\n");
		else methodeBuffer.append("Label" + label + ":\n");
	}

	public void appendLine(int line) {
		if (global) methodeBuffer.append(".line " + line + '\n');
		else methodeBuffer.append(".line " + line + '\n');
	}

	public void increaseStackBy(int stackdiff) {
		if (global) {
			clinitCurrentStack += stackdiff;
			if (clinitCurrentStack > clinitMaximumStack) clinitMaximumStack = clinitCurrentStack;
		}
		else {
			methodeCurrentStack += stackdiff;
			if (methodeCurrentStack > methodeMaximumStack) methodeMaximumStack = methodeCurrentStack;
		}
	}

	public void addLocalVariable() {
		methodeNumberOfLocals++;
	}

	public void addGlobalVariable(String name, int typ) {
		staticBuffer.append(".field static " + name);
		if (typ == Minako.TYP_INT) staticBuffer.append(" I\n");
		else if (typ == Minako.TYP_FLOAT) staticBuffer.append(" F\n");
		else if (typ == Minako.TYP_BOOLEAN) staticBuffer.append(" Z\n");
		else assert(false);
	}
}
